@@ -3,7 +3,7 @@
     pyckmeans, a Python package for Consensus K-Means clustering.
 '''
 
-__version__ = '0.5.0'
+__version__ = '0.6.0'
 
 __all__ = [
     'CKmeans',
@@ -13,7 +13,7 @@ __all__ = [
     'pcoa',
 ]
 
-from .core import CKmeans, MultiCKMeans
+from .core import CKmeans, MultiCKMeans, WECR
 from .io import NucleotideAlignment
 from .distance import DistanceMatrix
 from .ordination import pcoa
